#ifndef _MATRIX_HH
#define _MATRIX_HH

#undef EXPENSIVE_ASSERT
#define BOOST_UBLAS_TYPE_CHECK 0
#include "FastRationals.hh"
#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector_sparse.hpp>

typedef FastInteger ZZ;
typedef FastRational QQ;
typedef double FF;

using namespace boost::numeric::ublas;

#define SPARSE

#ifdef SPARSE
#define MY_MATRIX boost::numeric::ublas::compressed_matrix
#define MY_VECTOR boost::numeric::ublas::compressed_vector
#else
#define MY_MATRIX boost::numeric::ublas::matrix
#define MY_VECTOR boost::numeric::ublas::vector
#endif

typedef MY_MATRIX<QQ> qq_matrix;
typedef MY_VECTOR<QQ> qq_vector;
typedef MY_MATRIX<FF> ff_matrix;
typedef MY_VECTOR<FF> ff_vector;

typedef unsigned mat_index;

bool is_zero(const qq_matrix& m);

std::ostream& operator<<(std::ostream& o, const qq_matrix& m);
std::ostream& operator<<(std::ostream& o, const qq_vector& m);
std::ostream& operator<<(std::ostream& o, const ff_matrix& m);
std::ostream& operator<<(std::ostream& o, const ff_vector& m);

mat_index echelonize(qq_matrix& m, std::vector<mat_index>* pivot_columns=0);
qq_matrix right_kernel(const qq_matrix& m_src);
qq_matrix echelonize_and_right_kernel(qq_matrix& m);
qq_matrix build_from_rows(const std::vector<qq_vector>& rows);

class generator_system {
  bool empty_flag;
  qq_vector f0;
  qq_matrix fi;

public:
  generator_system(mat_index dimension);
  generator_system(mat_index nr_vectors, mat_index dimension);
  generator_system(const qq_vector& f0, const std::vector<qq_vector>& fi);

  void constrain(const qq_matrix& constraints);

  void print(std::ostream& o) const;

  mat_index ambient_dimension() const {
    return f0.size();
  }

  mat_index dimension() const {
    if (empty_flag) return -1;
    else return fi.size1();
  }

  void set_origin_coeff(unsigned j, const QQ& value) {
    f0(j) = value;
  }

  void set_vector_coeff(unsigned i, unsigned j, const QQ& value) {
    fi(i, j) = value;
  }

  QQ get_origin_coeff(unsigned j) const {
    return f0(j);
  }

  QQ get_vector_coeff(unsigned i, unsigned j) const {
    return fi(i, j);
  }

  const qq_vector& get_f0() const {
    return f0;
  }

  qq_vector get_fi(unsigned i) const {
    assert(i < fi.size1());
    return row(fi, i);
  }

  const qq_matrix get_fis() const {
    return fi;
  }

  qq_vector at(const qq_vector& y) const;
  ff_vector at(const ff_vector& y) const;

  bool is_empty() {
    return empty_flag;
  }

  ff_vector project(const ff_vector& v) const;
};

inline std::ostream& operator<<(std::ostream& o, const generator_system& m) {
  m.print(o);
  return o;
}

QQ rationalize(FF x, double epsilon=1e-7, int depth=10);

typedef std::vector<std::pair<QQ, qq_vector> > sum_of_squares;

struct Is_not_sdp {
};

sum_of_squares sos_decompose(qq_matrix m) throw(Is_not_sdp);
qq_matrix sos_recompose(unsigned dim, const sum_of_squares& sos);
bool is_sdp(qq_matrix m);
bool is_dp(qq_matrix m);

ff_vector qq_to_ff_vector(const qq_vector& v);
ff_matrix qq_to_ff_matrix(const qq_matrix& m);
#endif
