sparseflag=True
load 'quadratic_forms.sage'
load 'sdp_solver.sage'
load 'linear_algebra.sage'

### Polynomials
indeterminates=['x', 'y', 'z']
polynomial_ring=PolynomialRing(QQ, indeterminates)

monomial_list=[polynomial_ring('x')^a * polynomial_ring('y')^b * polynomial_ring('z')^c for a in range(3) for b in range(3) for c in range(3) if a+b+c<=2]

lhs_polynomials=[polynomial_ring('x*y + 3*y^2 + z + x^3 + 1'), polynomial_ring('x - 2*y^2 + 5*z^3 + 2'), polynomial_ring('x^2 - z + y'), 1]
lhs_polynomials.append(-(lhs_polynomials[0]+polynomial_ring('3+(x+5*y)^2')*lhs_polynomials[1]+lhs_polynomials[2])-polynomial_ring('1+x^2'));

lhs_monomial_lists=[monomial_list for p in lhs_polynomials]
assert len(lhs_polynomials) == len(lhs_monomial_lists)

#rhs_polynomial=polynomial_ring('3/4*(5*x+3*y^2+z^2)^2+2*(5*x+3*y^2+z^2)+5+(11*x+2*y+z)^4')
# rhs_polynomial=polynomial_ring('3/4*(551*x^2+3*y^2+z)^2+2*(551*x^2+3*y^2+z)+5+(23*x+13*y+z)^2')
rhs_polynomial=polynomial_ring(-1)
#rhs_polynomial=polynomial_ring('3/4*(551*x^2+3*y^2+z)^2+2*(551*x^2+3*y^2+z)+5+(23*x+13*y+z)^2')
# rhs_polynomial=polynomial_ring('3/4*(x+3*y^2+z)^2+2*(x+3*y^2+z)+5')
#rhs_polynomial=polynomial_ring('3/4*(x+3*y^2+z)^2+2*(x+3*y^2+z)+5+(1+x+y+z)^2')


def coeff_name(block, i, j):
  if i<j:
    k=i
    i=j
    j=k
  return "mat_%d_%d_%d" %(block, i, j)

coeff_positions=[(block, i, j)\
             for block in xrange(len(lhs_monomial_lists))\
             for i in xrange(len(lhs_monomial_lists[block]))\
             for j in xrange(i+1)]
coeff_positions_rev = dict([(coeff_positions[k], k) for k in xrange(len(coeff_positions))])

coeff_names=[coeff_name(block, i, j) for (block, i, j) in coeff_positions]

def get_coeff_position(block, i, j):
  if i < j:
    return coeff_positions_rev[(block, j, i)]
  else:
    return coeff_positions_rev[(block, i, j)]
  
def vector_to_blocks(v):
  blocks = [matrix(QQ, len(block), sparse=sparseflag) for block in lhs_monomial_lists]
  for k in v.nonzero_positions():
    (block, i, j)=coeff_positions[k]
    blocks[block][i,j] = blocks[block][j,i] = v[k]
  return blocks

def polynomial_from_blocks(blocks):
  total = polynomial_ring(0)
  for i in xrange(len(lhs_polynomials)):
    total += apply_qmatrix(blocks[i], vector(lhs_monomial_lists[i])) * lhs_polynomials[i]
  return total

def apply_solution(solution, zf0b, zfisb):
  assert len(solution) == len(zfisb)
  result = copy(zf0b)
  for block in xrange(len(result)):
    for i in xrange(len(solution)):
      result[block] += solution[i] * zfisb[i][block]
  return result

constraint_space = VectorSpace(QQ, len(coeff_names)+1, sparse=sparseflag)
constraint_map = {}
for (exponents_t, scalar) in rhs_polynomial.dict().iteritems():
  exponents = tuple(exponents_t)
  v=constraint_space(0)
  v[len(v)-1]=-scalar
  constraint_map[exponents]=v

for block in xrange(len(lhs_polynomials)):
  monomials = lhs_monomial_lists[block]
  for i in xrange(len(monomials)):
    for j in xrange(len(monomials)):
      k=get_coeff_position(block, i, j)
      for (exponents_t, coeff) in (monomials[i]*monomials[j]*lhs_polynomials[block]).dict().iteritems():
        exponents=tuple(exponents_t)
        if not constraint_map.has_key(exponents):
          constraint_map[exponents]=constraint_space(0)
        constraint_map[exponents][k] += coeff

constraint_matrix = build_from_rows(constraint_map.values())
print "echelonizing"
constraint_matrix.echelonize();
    
nonzero_rows = constraint_matrix.pivot_rows();
if len(nonzero_rows)==0:
  assert false
last_nonzero_row=nonzero_rows[len(nonzero_rows)-1]
if last_nonzero_row < constraint_matrix.nrows()-1:
  constraint_matrix=constraint_matrix[0:last_nonzero_row+1,:]

print "echelonized, computing kernel"
print "constraints:", constraint_matrix.nrows()
constraint_solution = right_kernel_matrix_from_echelon(constraint_matrix)

fis=[]
f0=()
for row in constraint_solution.rows():
  if row[len(row)-1]==0:
    fis.append(row[0:len(row)-1])
  else:
    if f0==():
      f0=row[0:len(row)-1]/row[len(row)-1]
    else:
      assert false
          
if f0==():
  has_solution=False

print "dimension of search space:", len(fis)

print 'conversion to blocks'
f0b = vector_to_blocks(f0)
fisb = map(vector_to_blocks, fis)

assert(polynomial_from_blocks(f0b) == rhs_polynomial)
for fib in fisb:
  assert(polynomial_from_blocks([f0b[block] + fib[block] for block in xrange(len(f0b))]) == rhs_polynomial)
  
print 'solve sdp'
matrices=sdp_solve_q("sosproblem", [-m for m in f0b], fisb, [0 for fib in fisb])
if matrices==():
  print "NO SOLUTION"
else:
  assert(polynomial_from_blocks(matrices) == rhs_polynomial)

  total=polynomial_ring(0)
  for block in xrange(len(lhs_polynomials)):
    print "For: ", lhs_polynomials[block]
    sos=polynomial_ring(0)
    for (coeff, v) in sdp_matrix_to_sos(matrices[block]):
      print "  ", coeff, "(", (v*vector(lhs_monomial_lists[block])), ")^2"
      sos += coeff * (v*vector(lhs_monomial_lists[block]))^2
    total += lhs_polynomials[block] * sos

  print "SOLUTION IS CORRECT:", total == rhs_polynomial
